from dataclasses import MISSING, Field, fields
from datetime import date
from re import fullmatch
from typing import ClassVar

from sqlalchemy.orm import DeclarativeBase, Mapped, MappedAsDataclass, mapped_column

from .custom_types import int_pk, samplesheet_str_pk, stripped_str_pk

__all__ = ['Base', 'Entity', 'Data', 'Process']


class Base(MappedAsDataclass, DeclarativeBase):
    @classmethod
    def dc_fields(cls) -> tuple[Field, ...]:
        return fields(cls)

    @classmethod
    def dc_init_fields(cls) -> tuple[Field, ...]:
        return tuple(f for f in cls.dc_fields() if f.init)

    @classmethod
    def required_dc_init_fields(cls) -> tuple[Field, ...]:
        return tuple(
            f
            for f in fields(cls)
            if f.init and f.default is MISSING and f.default_factory is MISSING
        )

    @classmethod
    def dc_field_names(cls) -> set[str]:
        return {f.name for f in cls.dc_fields()}

    @classmethod
    def dc_init_field_names(cls) -> set[str]:
        return {f.name for f in cls.dc_init_fields()}

    @classmethod
    def required_dc_init_field_names(cls) -> set[str]:
        return {f.name for f in cls.required_dc_init_fields()}

    pass


class Entity(Base, kw_only=True):
    __abstract__ = True


class Data(Base, kw_only=True):
    __abstract__ = True

    # TODO: auto incrementing behavior
    id: Mapped[samplesheet_str_pk]

    # Model metadata
    id_date_col: ClassVar[str]
    id_prefix: ClassVar[str]
    id_length: ClassVar[int]

    def __post_init__(self):
        self.id = self.id.strip().upper()

        date_col: date = getattr(self, self.id_date_col)

        if date_col is None:
            prefix = self.id_prefix
            suffix_length = self.id_length - len(self.id_prefix)
        else:
            year_last_two_digits = date_col.strftime('%y')
            year_last_two_digits_as_int = int(date_col.strftime('%y'))

            year_pattern = '|'.join(
                str(year_last_two_digits_as_int + i) for i in range(-1, 2)
            )

            prefix = self.id_prefix + f'({year_pattern})'
            suffix_length = (
                self.id_length - len(self.id_prefix) - len(year_last_two_digits)
            )

        pattern = rf'{prefix}\d{{{suffix_length}}}'

        model_name = type(self).__name__

        if fullmatch(pattern, self.id) is None:
            raise ValueError(
                f'{model_name} ID {self.id} does not match the pattern {pattern}.'
            )


class Process(Base, kw_only=True):
    __abstract__ = True

    name: Mapped[stripped_str_pk]



from functools import cached_property
from os import environ
from pathlib import Path
from re import fullmatch

from email_validator import validate_email
from sqlalchemy import ForeignKey, UniqueConstraint, inspect
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from urllib3 import request

from ..bases import Entity
from ..custom_types import (
    SamplesheetString,
    StrippedString,
    stripped_str,
    stripped_str_pk,
    unique_stripped_str,
)
from ..utils import get_format_string_vars

__all__ = ['Institution', 'Lab', 'Person']


class Institution(Entity, kw_only=True):
    __tablename__ = 'institution'

    # Institution attributes
    name: Mapped[stripped_str_pk] = mapped_column(default=None)
    email_format: Mapped[stripped_str] = mapped_column(repr=False)
    ror_id: Mapped[unique_stripped_str | None] = mapped_column(default=None, repr=False)
    short_name: Mapped[stripped_str] = mapped_column(
        default=None, index=True, repr=False
    )
    country: Mapped[str] = mapped_column(
        StrippedString(length=2), default='US', repr=False
    )
    state: Mapped[str | None] = mapped_column(
        StrippedString(length=2), default=None, repr=False
    )
    city: Mapped[stripped_str] = mapped_column(default=None, repr=False)

    @validates('email_format')
    def validate_email_format(self, key: str, email_format: str) -> str:
        email_format = email_format.strip().lower()

        variables = get_format_string_vars(email_format)

        if not variables:
            raise ValueError(f'No variables found in {key} {email_format}')

        person_columns = set(inspect(Person).columns.keys())
        non_existent_person_columns = variables - person_columns

        if non_existent_person_columns:
            raise ValueError(
                f'The following variables in {email_format} are not members of {person_columns}:\n{non_existent_person_columns}'
            )

        example_values = {var: 'string' for var in variables}
        example_email = email_format.format_map(example_values)

        validate_email(example_email)

        return email_format

    @cached_property
    def _ror_data(self):
        if self.ror_id is None:
            return None

        base_url = 'https://api.ror.org/organizations'
        url = f'{base_url}/{self.ror_id}'
        response = request(method='GET', url=url)

        if response.status != 200:
            raise ValueError(
                f'ROR ID {self.ror_id} not found in database search of {base_url}'
            )

        return response.json()

    def __post_init__(self):
        if self._ror_data is None:
            if not all((self.name, self.short_name, self.country, self.city)):
                raise ValueError(
                    'If ROR ID is not provided, then name, short_name, country, and city must be provided.'
                )
            return

        if self.name is None:
            self.name = self._ror_data['name']

        acronyms = self._ror_data['acronyms']
        aliases = self._ror_data['aliases']
        if self.short_name is None:
            if len(acronyms) > 0:
                self.short_name = acronyms[0]
            elif len(aliases) > 0:
                self.short_name = aliases[0]
            else:
                raise ValueError(
                    'Could not find short name from ROR data. Please provide manually.'
                )

        self.country = self._ror_data['country']['country_code']

        acronyms = self._ror_data['acronyms']
        aliases = self._ror_data['aliases']
        if self.short_name is None:
            if len(acronyms) > 0:
                self.short_name = acronyms[0]
            elif len(aliases) > 0:
                self.short_name = aliases[0]
            else:
                pass

        self.country = self._ror_data['country']['country_code']

        addresses = self._ror_data['addresses']
        if len(addresses) > 0:
            geonames_city_info = addresses[0]['geonames_city']
            self.city = geonames_city_info['city']
            if self.country == 'US':
                state_code = geonames_city_info['geonames_admin1']['code']
                _, self.state = state_code.split('.')
        else:
            raise ValueError(
                f'Could not find city information from ROR for {self.name}. Please enter manually.'
            )


class Person(Entity, kw_only=True):
    __tablename__ = 'person'

    # Person attributes
    first_name: Mapped[stripped_str]
    last_name: Mapped[stripped_str]
    orcid: Mapped[unique_stripped_str | None] = mapped_column(default=None, repr=False)
    email_auto_generated: Mapped[bool] = mapped_column(
        init=False, default=False, repr=False
    )
    email: Mapped[stripped_str_pk] = mapped_column(default=None)

    # Parent foreign keys
    institution_name: Mapped[str] = mapped_column(
        ForeignKey('institution.name'), init=False, repr=False
    )

    # Parent models
    institution: Mapped[Institution] = relationship(repr=False)

    @validates('first_name', 'last_name')
    def format_name(self, key: str, name: str) -> str:
        formatted_split = name.strip().title().split()
        normalized_inner_whitespace = ' '.join(formatted_split)
        return normalized_inner_whitespace

    @validates('orcid')
    def validate_orcid(self, key: str, orcid: str | None) -> str | None:
        if orcid is None:
            return orcid

        orcid = orcid.strip()

        orcid_pattern = r'(\d{4})-?(\d{4})-?(\d{4})-?(\d{4}|\d{3}X)'
        if (match_obj := fullmatch(orcid_pattern, string=orcid)) is None:
            raise ValueError(f'ORCID {orcid} does not match pattern {orcid_pattern}')

        digit_groups = match_obj.groups()
        formatted_orcid = '-'.join(digit_groups)

        base_url = 'https://pub.orcid.org'
        url = f'{base_url}/{formatted_orcid}'
        headers = {'Accept': 'application/json'}
        response = request(method='GET', url=url, headers=headers)

        if response.status != 200:
            raise ValueError(
                f'{formatted_orcid} not found in database search of {base_url}'
            )

        return formatted_orcid

    @validates('email')
    def validate_email(self, key: str, email: str | None) -> str | None:
        if email is None:
            return email

        email_info = validate_email(email.lower(), check_deliverability=True)
        return email_info.normalized

    def __post_init__(self):
        if self.email is not None:
            return

        variables = get_format_string_vars(self.institution.email_format)
        var_values = {var: getattr(self, var) for var in variables}

        self.email = self.institution.email_format.format_map(var_values).replace(
            ' ', ''
        )
        self.email_auto_generated = True


class Lab(Entity, kw_only=True):
    __tablename__ = 'lab'

    # Lab attributes
    name: Mapped[stripped_str_pk] = mapped_column(default=None)
    delivery_dir: Mapped[unique_stripped_str] = mapped_column(default=None, repr=False)
    unix_group: Mapped[stripped_str] = mapped_column(
        init=False, default=None, compare=False, repr=False
    )

    # Parent foreign keys
    institution_name: Mapped[int] = mapped_column(
        ForeignKey('institution.name'),
        repr=False,
        compare=False,
        init=False,
        primary_key=True,
    )
    pi_email: Mapped[str] = mapped_column(
        ForeignKey('person.email'),
        compare=False,
        init=False,
        primary_key=True,
        repr=False,
    )

    # Parent models
    institution: Mapped[Institution] = relationship(repr=False)
    pi: Mapped[Person] = relationship(repr=False)

    @cached_property
    def _delivery_parent_dir(self) -> Path:
        return Path(environ['delivery_parent_dir'])

    @validates('delivery_dir')
    def validate_delivery_dir(self, key: str, delivery_dir: str | None):
        if delivery_dir is None:
            return delivery_dir

        delivery_path = self._delivery_parent_dir / delivery_dir
        if not delivery_path.is_dir():
            raise NotADirectoryError(f'{key} {delivery_path} is not a directory.')

        return str(delivery_path)

    def __post_init__(self):
        if self.name is None:
            self.name = f'{self.pi.first_name} {self.pi.last_name} Lab'

        if self.delivery_dir is not None:
            self.unix_group = Path(self.delivery_dir).group()
            return

        first_name = SamplesheetString().process_bind_param(
            self.pi.first_name.lower(), dialect=None
        )
        last_name = SamplesheetString().process_bind_param(
            self.pi.last_name.lower(), dialect=None
        )

        delivery_path = self._delivery_parent_dir / f'{first_name}_{last_name}'
        if not delivery_path.is_dir():
            raise NotADirectoryError(
                f'Delivery directory {delivery_path} does not exist.'
            )

        self.delivery_dir = str(delivery_path)
        self.unix_group = delivery_path.group()



from datetime import date
from re import fullmatch
from typing import ClassVar, Literal

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ...bases import Base, Data, Process
from ...custom_types import samplesheet_str, samplesheet_str_pk, stripped_str
from ..data import DataSet, Sample

__all__ = [
    'ChromiumDataSet',
    'ChromiumTag',
    'ChromiumSample',
    'SequencingRun',
    'ChromiumLibraryType',
    'ChromiumLibrary',
]


class ChromiumDataSet(DataSet, kw_only=True):
    # Child models
    samples: Mapped[list['ChromiumSample']] = relationship(
        back_populates='data_set', default_factory=list, repr=False
    )
    libraries: Mapped[list['ChromiumLibrary']] = relationship(
        back_populates='data_set', default_factory=list, repr=False
    )

    # Model metadata
    id_prefix: ClassVar[Literal['SD']] = 'SD'
    id_length: ClassVar[Literal[9]] = 9

    __mapper_args__ = {
        'polymorphic_identity': 'Chromium',
    }


class ChromiumTag(Base, kw_only=True):
    __tablename__ = 'chromium_tag'

    # TODO: add validation
    # Tag attributes
    id: Mapped[samplesheet_str_pk]
    name: Mapped[samplesheet_str | None]
    type: Mapped[stripped_str]
    read: Mapped[stripped_str]
    sequence: Mapped[stripped_str]
    pattern: Mapped[stripped_str]
    five_prime_offset: Mapped[int]


class ChromiumSample(Sample, kw_only=True):
    # Sample attributes
    targeted_cell_recovery: Mapped[int | None] = mapped_column(default=None)

    # Parent foreign keys
    tag_id: Mapped[str | None] = mapped_column(
        ForeignKey('chromium_tag.id'), default=None, repr=False, init=False
    )

    # Parent models
    data_set: Mapped[ChromiumDataSet] = relationship(
        back_populates='samples', default=None, repr=False
    )
    tag: Mapped[ChromiumTag | None] = relationship(default=None, repr=False)

    # Model metadata
    id_prefix: ClassVar[Literal['SS']] = 'SS'
    id_length: ClassVar[Literal[9]] = 9

    __mapper_args__ = {'polymorphic_identity': 'Chromium'}


class SequencingRun(Data, kw_only=True):
    __tablename__ = 'sequencing_run'

    libraries: Mapped[list['ChromiumLibrary']] = relationship(
        back_populates='sequencing_run', default_factory=list, repr=False
    )

    def __post_init__(self):
        self.id = self.id.strip().lower()

        pattern = rf'\d{{2}}-scbct-\d{{2,3}}'

        model_name = type(self).__name__
        if fullmatch(pattern, self.id) is None:
            raise ValueError(
                f'{model_name} ID {self.id} does not match the pattern {pattern}.'
            )


class ChromiumLibraryType(Process, kw_only=True):
    __tablename__ = 'chromium_library_type'


class ChromiumLibrary(Data, kw_only=True):
    __tablename__ = 'chromium_library'

    # Library attributes
    date_constructed: Mapped[date | None] = mapped_column(default=None, repr=False)

    # TODO: add some validation so that libraries with a particular
    # status must have a sequencing run
    status: Mapped[stripped_str | None] = mapped_column(default=None, repr=False)

    # Parent foreign keys
    data_set_id: Mapped[str] = mapped_column(
        ForeignKey('data_set.id'), default=None, init=False, repr=False
    )
    library_type_name: Mapped[str] = mapped_column(
        ForeignKey('chromium_library_type.name'), default=None, init=False, repr=False
    )
    sequencing_run_id: Mapped[str | None] = mapped_column(
        ForeignKey('sequencing_run.id'),
        default=None,
        compare=False,
        init=False,
        repr=False,
    )

    # Parent models
    data_set: Mapped[ChromiumDataSet] = relationship(
        back_populates='libraries', default=None, repr=False
    )
    library_type: Mapped[ChromiumLibraryType] = relationship(repr=False)
    sequencing_run: Mapped[SequencingRun | None] = relationship(
        default=None, compare=False, repr=False
    )

    # Model metadata
    id_date_col: ClassVar[Literal['date_constructed']] = 'date_constructed'
    id_prefix: ClassVar[Literal['SC']] = 'SC'
    id_length: ClassVar[Literal[9]] = 9



from .bases import Base
from .models.entities import *
from .models.platforms.chromium import *
from .models.platforms.xenium import *
from .models.processes import *

ORDERED_MODELS: dict[str, type[Base]] = {
    'Institution': Institution,
    'Person': Person,
    'Lab': Lab,
    'Platform': Platform,
    'Assay': Assay,
    'SequencingRun': SequencingRun,
    'ChromiumDataSet': ChromiumDataSet,
    'ChromiumTag': ChromiumTag,
    'ChromiumSample': ChromiumSample,
    'ChromiumLibraryType': ChromiumLibraryType,
    'ChromiumLibrary': ChromiumLibrary,
    'XeniumRun': XeniumRun,
    'XeniumDataSet': XeniumDataSet,
    'XeniumRegion': XeniumRegion,
}
__all__ = [
    'Base',
    'Institution',
    'Person',
    'Lab',
    'Platform',
    'Assay',
    'ChromiumDataSet',
    'ChromiumTag',
    'ChromiumSample',
    'SequencingRun',
    'ChromiumLibraryType',
    'ChromiumLibrary',
    'XeniumRun',
    'XeniumDataSet',
    'XeniumRegion',
    'ORDERED_MODELS',
]
